@@ -1,5 +1,5 @@
 /*  Tarlz - Archiver with multimember lzip compression
-    Copyright (C) 2013-2018 Antonio Diaz Diaz.
+    Copyright (C) 2013-2019 Antonio Diaz Diaz.
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -18,39 +18,44 @@
 #define _FILE_OFFSET_BITS 64
 
 #include <algorithm>
+#include <cctype>
 #include <cerrno>
+#include <climits>
 #include <cstdio>
 #include <cstdlib>
 #include <cstring>
 #include <string>
 #include <vector>
+#include <pthread.h>
 #include <stdint.h>
 #include <unistd.h>
 #include <utime.h>
 #include <sys/stat.h>
 #include <sys/types.h>
 #if defined(__GNU_LIBRARY__)
-#include <sys/sysmacros.h>		// for makedev
+#include <sys/sysmacros.h>		// for major, minor, makedev
 #endif
 #include <lzlib.h>
 
 #include "arg_parser.h"
-#include "lzip.h"
+#include "lzip_index.h"
 #include "tarlz.h"
 
 
 namespace {
 
-int gretval = 0;
-bool has_lz_ext;			// global var for archive_read
+Resizable_buffer grbuf( initial_line_length );
+bool archive_is_uncompressed_seekable = false;
+bool has_lz_ext;				// global var for archive_read
 
-void skip_warn( const bool reset = false )	// avoid duplicate warnings
+bool skip_warn( const bool reset = false )	// avoid duplicate warnings
   {
   static bool skipping = false;
 
   if( reset ) skipping = false;
   else if( !skipping )
-    { skipping = true; show_error( "Skipping to next header." ); }
+    { skipping = true; show_error( "Skipping to next header." ); return true; }
+  return false;
   }
 
 
@@ -83,13 +88,6 @@ bool make_path( const std::string & name )
   }
 
 
-inline bool block_is_zero( const uint8_t * const buf, const int size )
-  {
-  for( int i = 0; i < size; ++i ) if( buf[i] != 0 ) return false;
-  return true;
-  }
-
-
 // Return value: 0 = OK, 1 = damaged member, 2 = fatal error.
 // If sizep and error, return in *sizep the number of bytes read.
 // The first 6 bytes of the archive must be intact for islz to be meaningful.
@@ -114,6 +112,7 @@ int archive_read( const int infd, uint8_t * const buf, const int size,
       { show_error( "Error reading archive", errno ); fatal = true; return 2; }
     const Lzip_header & header = (*(const Lzip_header *)buf);
     bool islz = ( rd >= min_member_size && header.verify_magic() &&
+                  header.verify_version() &&
                   isvalid_ds( header.dictionary_size() ) );
     const bool istar = ( rd == size && verify_ustar_chksum( buf ) );
     const bool iseof =
@@ -121,16 +120,15 @@ int archive_read( const int infd, uint8_t * const buf, const int size,
     if( !islz && !istar && !iseof )		// corrupt or invalid format
       {
       show_error( "This does not look like a POSIX tar archive." );
-      if( has_lz_ext ) islz = true;
-      if( verbosity >= 2 && !islz && rd == size )
-        std::fprintf( stderr, "ustar chksum = %07o\n", ustar_chksum( buf ) );
+      if( has_lz_ext && rd >= min_member_size ) islz = true;
       if( !islz ) return 1;
       }
     if( !islz )						// uncompressed
       { if( rd == size ) return 0; fatal = true; return 2; }
-    decoder = LZ_decompress_open();			// compressed
+    archive_is_uncompressed_seekable = false;		// compressed
+    decoder = LZ_decompress_open();
     if( !decoder || LZ_decompress_errno( decoder ) != LZ_ok )
-      { show_error( "Not enough memory." );
+      { show_error( mem_msg );
         LZ_decompress_close( decoder ); fatal = true; return 2; }
     if( LZ_decompress_write( decoder, buf, rd ) != rd )
       internal_error( "library error (LZ_decompress_write)." );
@@ -157,11 +155,11 @@ int archive_read( const int infd, uint8_t * const buf, const int size,
       {
       if( LZ_decompress_sync_to_member( decoder ) < 0 )
         internal_error( "library error (LZ_decompress_sync_to_member)." );
-      skip_warn(); gretval = 2; return 1;
+      skip_warn(); set_error_status( 2 ); return 1;
       }
     if( rd == 0 && LZ_decompress_finished( decoder ) == 1 )
-     { LZ_decompress_close( decoder );
-       show_error( "Archive ends unexpectedly." ); fatal = true; return 2; }
+      { LZ_decompress_close( decoder );
+        show_error( "Archive ends unexpectedly." ); fatal = true; return 2; }
     sz += rd; if( sizep ) *sizep = sz;
     if( sz == size && LZ_decompress_finished( decoder ) == 1 &&
         LZ_decompress_close( decoder ) < 0 )
@@ -185,12 +183,14 @@ int archive_read( const int infd, uint8_t * const buf, const int size,
   }
 
 
-const char * mode_string( const Tar_header header )
+enum { mode_string_size = 10,
+       group_string_size = 1 + uname_l + 1 + gname_l + 1 };	// 67
+
+void format_mode_string( const Tar_header header, char buf[mode_string_size] )
   {
-  static char buf[11];
   const Typeflag typeflag = (Typeflag)header[typeflag_o];
 
-  std::memcpy( buf, "----------", sizeof buf - 1 );
+  std::memcpy( buf, "----------", mode_string_size );
   switch( typeflag )
     {
     case tf_regular: break;
@@ -203,7 +203,7 @@ const char * mode_string( const Tar_header header )
     case tf_hiperf: buf[0] = 'C'; break;
     default: buf[0] = '?';
     }
-  const mode_t mode = strtoul( header + mode_o, 0, 8 );		// 12 bits
+  const mode_t mode = parse_octal( header + mode_o, mode_l );	// 12 bits
   const bool setuid = mode & S_ISUID;
   const bool setgid = mode & S_ISGID;
   const bool sticky = mode & S_ISVTX;
@@ -219,60 +219,99 @@ const char * mode_string( const Tar_header header )
   if( mode & S_IWOTH ) buf[8] = 'w';
   if( mode & S_IXOTH ) buf[9] = sticky ? 't' : 'x';
   else if( sticky ) buf[9] = 'T';
-  return buf;
   }
 
 
-const char * user_group_string( const Tar_header header )
+int format_user_group_string( const Tar_header header,
+                              char buf[group_string_size] )
   {
-  enum { bufsize = uname_l + 1 + gname_l + 1 };
-  static char buf[bufsize];
-
+  int len;
   if( header[uname_o] && header[gname_o] )
-    snprintf( buf, bufsize, "%.32s/%.32s", header + uname_o, header + gname_o );
+    len = snprintf( buf, group_string_size,
+                    " %.32s/%.32s", header + uname_o, header + gname_o );
   else
     {
-    const int uid = strtoul( header + uid_o, 0, 8 );
-    const int gid = strtoul( header + gid_o, 0, 8 );
-    snprintf( buf, bufsize, "%u/%u", uid, gid );
+    const unsigned uid = parse_octal( header + uid_o, uid_l );
+    const unsigned gid = parse_octal( header + gid_o, gid_l );
+    len = snprintf( buf, group_string_size, " %u/%u", uid, gid );
     }
-  return buf;
+  return len;
+  }
+
+} // end namespace
+
+bool block_is_zero( const uint8_t * const buf, const int size )
+  {
+  for( int i = 0; i < size; ++i ) if( buf[i] != 0 ) return false;
+  return true;
   }
 
 
-void show_member_name( const Extended & extended, const Tar_header header,
-                       const int vlevel )
+void format_member_name( const Extended & extended, const Tar_header header,
+                         Resizable_buffer & rbuf, const bool long_format )
   {
-  if( verbosity < vlevel ) return;
-  if( verbosity > vlevel )
+  if( long_format )
     {
-    const time_t mtime = strtoull( header + mtime_o, 0, 8 );	// 33 bits
-    const struct tm * const tm = localtime( &mtime );
+    format_mode_string( header, rbuf() );
+    const int group_string_len =
+      format_user_group_string( header, rbuf() + mode_string_size );
+    int offset = mode_string_size + group_string_len;
+    const time_t mtime = parse_octal( header + mtime_o, mtime_l ); // 33 bits
+    struct tm tms;
+    const struct tm * tm = localtime_r( &mtime, &tms );
+    if( !tm )
+      { time_t z = 0; tm = localtime_r( &z, &tms ); if( !tm ) tm = &tms; }
     const Typeflag typeflag = (Typeflag)header[typeflag_o];
     const bool islink = ( typeflag == tf_link || typeflag == tf_symlink );
     const char * const link_string = !islink ? "" :
                          ( ( typeflag == tf_link ) ? " link to " : " -> " );
-    std::printf( "%s %s %9llu %4d-%02u-%02u %02u:%02u %s%s%s\n",
-                 mode_string( header ), user_group_string( header ),
-                 extended.size, 1900 + tm->tm_year, 1 + tm->tm_mon,
-                 tm->tm_mday, tm->tm_hour, tm->tm_min, extended.path.c_str(),
-                 link_string, !islink ? "" : extended.linkpath.c_str() );
+    if( typeflag == tf_chardev || typeflag == tf_blockdev )
+      offset += snprintf( rbuf() + offset, rbuf.size() - offset, " %5u,%u",
+                  (unsigned)parse_octal( header + devmajor_o, devmajor_l ),
+                  (unsigned)parse_octal( header + devminor_o, devminor_l ) );
+    else
+      offset += snprintf( rbuf() + offset, rbuf.size() - offset, " %9llu",
+                          extended.file_size() );
+    for( int i = 0; i < 2; ++i )
+      {
+      const int len = snprintf( rbuf() + offset, rbuf.size() - offset,
+                " %4d-%02u-%02u %02u:%02u %s%s%s\n",
+                1900 + tm->tm_year, 1 + tm->tm_mon, tm->tm_mday,
+                tm->tm_hour, tm->tm_min, extended.path().c_str(),
+                link_string, islink ? extended.linkpath().c_str() : "" );
+      if( (int)rbuf.size() > len + offset || !rbuf.resize( len + offset + 1 ) )
+        break;
+      }
     }
-  else std::printf( "%s\n", extended.path.c_str() );
+  else
+    {
+    if( rbuf.size() < extended.path().size() + 2 )
+      rbuf.resize( extended.path().size() + 2 );
+    snprintf( rbuf(), rbuf.size(), "%s\n", extended.path().c_str() );
+    }
+  }
+
+namespace {
+
+void show_member_name( const Extended & extended, const Tar_header header,
+                       const int vlevel, Resizable_buffer & rbuf )
+  {
+  if( verbosity < vlevel ) return;
+  format_member_name( extended, header, rbuf, verbosity > vlevel );
+  std::fputs( rbuf(), stdout );
   std::fflush( stdout );
   }
 
 
-int list_member( const int infd, const Extended & extended,
-                 const Tar_header header, const bool skip )
+int skip_member( const int infd, const Extended & extended )
   {
-  if( !skip ) show_member_name( extended, header, 0 );
-
+  unsigned long long rest = extended.file_size();
+  const int rem = rest % header_size;
+  const int padding = rem ? header_size - rem : 0;
+  if( archive_is_uncompressed_seekable &&
+      lseek( infd, rest + padding, SEEK_CUR ) > 0 ) return 0;
   const unsigned bufsize = 32 * header_size;
   uint8_t buf[bufsize];
-  unsigned long long rest = extended.size;
-  const int rem = extended.size % header_size;
-  const int padding = rem ? header_size - rem : 0;
   while( rest > 0 )
     {
     const int rsize = ( rest >= bufsize ) ? bufsize : rest + padding;
@@ -285,6 +324,130 @@ int list_member( const int infd, const Extended & extended,
   }
 
 
+void show_file_diff( const char * const filename, const char * const msg )
+  {
+  if( verbosity >= 0 ) std::fprintf( stderr, "%s: %s\n", filename, msg );
+  }
+
+
+int compare_member( const int infd1, const Extended & extended,
+                    const Tar_header header, const bool ignore_ids )
+  {
+  show_member_name( extended, header, 1, grbuf );
+  unsigned long long rest = extended.file_size();
+  const char * const filename = extended.path().c_str();
+  const Typeflag typeflag = (Typeflag)header[typeflag_o];
+  bool diff = false, size_differs = false, type_differs = true;
+  struct stat st;
+  if( lstat( filename, &st ) != 0 )
+    show_file_error( filename, "Warning: Can't stat", errno );
+  else if( ( typeflag == tf_regular || typeflag == tf_hiperf ) &&
+           !S_ISREG( st.st_mode ) )
+    show_file_diff( filename, "Is not a regular file" );
+  else if( typeflag == tf_symlink && !S_ISLNK( st.st_mode ) )
+    show_file_diff( filename, "Is not a symlink" );
+  else if( typeflag == tf_chardev && !S_ISCHR( st.st_mode ) )
+    show_file_diff( filename, "Is not a character device" );
+  else if( typeflag == tf_blockdev && !S_ISBLK( st.st_mode ) )
+    show_file_diff( filename, "Is not a block device" );
+  else if( typeflag == tf_directory && !S_ISDIR( st.st_mode ) )
+    show_file_diff( filename, "Is not a directory" );
+  else if( typeflag == tf_fifo && !S_ISFIFO( st.st_mode ) )
+    show_file_diff( filename, "Is not a FIFO" );
+  else
+    {
+    type_differs = false;
+    if( typeflag != tf_symlink )
+      {
+      const mode_t mode = parse_octal( header + mode_o, mode_l );  // 12 bits
+      if( mode != ( st.st_mode & ( S_ISUID | S_ISGID | S_ISVTX |
+                                   S_IRWXU | S_IRWXG | S_IRWXO ) ) )
+        { show_file_diff( filename, "Mode differs" ); diff = true; }
+      }
+    if( !ignore_ids )
+      {
+      if( (uid_t)parse_octal( header + uid_o, uid_l ) != st.st_uid )
+        { show_file_diff( filename, "Uid differs" ); diff = true; }
+      if( (gid_t)parse_octal( header + gid_o, gid_l ) != st.st_gid )
+        { show_file_diff( filename, "Gid differs" ); diff = true; }
+      }
+    if( typeflag != tf_symlink )
+      {
+      const time_t mtime = parse_octal( header + mtime_o, mtime_l ); // 33 bits
+      if( mtime != st.st_mtime )
+        { show_file_diff( filename, "Mod time differs" ); diff = true; }
+      if( ( typeflag == tf_regular || typeflag == tf_hiperf ) &&
+          (off_t)rest != st.st_size )		// don't compare contents
+        { show_file_diff( filename, "Size differs" ); size_differs = true; }
+      if( ( typeflag == tf_chardev || typeflag == tf_blockdev ) &&
+          ( parse_octal( header + devmajor_o, devmajor_l ) != major( st.st_rdev ) ||
+            parse_octal( header + devminor_o, devminor_l ) != minor( st.st_rdev ) ) )
+        { show_file_diff( filename, "Device number differs" ); diff = true; }
+      }
+    else
+      {
+      char * const buf = new char[st.st_size+1];
+      long len = readlink( filename, buf, st.st_size );
+      bool e = ( len != st.st_size );
+      if( !e ) { buf[len] = 0; if( extended.linkpath() != buf ) e = true; }
+      delete[] buf;
+      if( e ) { show_file_diff( filename, "Symlink differs" ); diff = true; }
+      }
+    }
+  if( diff || size_differs || type_differs )
+    { diff = false; set_error_status( 1 ); }
+  if( rest == 0 ) return 0;
+  if( ( typeflag != tf_regular && typeflag != tf_hiperf ) ||
+      size_differs || type_differs ) return skip_member( infd1, extended );
+  // else compare file contents
+  const int rem = rest % header_size;
+  const int padding = rem ? header_size - rem : 0;
+  const unsigned bufsize = 32 * header_size;
+  uint8_t buf1[bufsize];
+  uint8_t buf2[bufsize];
+  const int infd2 = open_instream( filename );
+  if( infd2 < 0 )
+    { set_error_status( 1 ); return skip_member( infd1, extended ); }
+  int retval = 0;
+  while( rest > 0 )
+    {
+    const int rsize1 = ( rest >= bufsize ) ? bufsize : rest + padding;
+    const int rsize2 = ( rest >= bufsize ) ? bufsize : rest;
+    const int ret = archive_read( infd1, buf1, rsize1 );
+    if( ret != 0 ) { if( ret == 2 ) retval = 2; diff = true; break; }
+    if( !diff )
+      {
+      const int rd = readblock( infd2, buf2, rsize2 );
+      if( rd != rsize2 )
+        {
+        if( errno ) show_file_error( filename, "Read error", errno );
+        else show_file_diff( filename, "EOF found in file" );
+        diff = true;
+        }
+      else
+        {
+        int i = 0; while( i < rsize2 && buf1[i] == buf2[i] ) ++i;
+        if( i < rsize2 )
+          { show_file_diff( filename, "Contents differ" ); diff = true; }
+        }
+      }
+    if( rest < bufsize ) break;
+    rest -= rsize1;
+    }
+  if( diff ) set_error_status( 1 );
+  close( infd2 );
+  return retval;
+  }
+
+
+int list_member( const int infd, const Extended & extended,
+                 const Tar_header header )
+  {
+  show_member_name( extended, header, 0, grbuf );
+  return skip_member( infd, extended );
+  }
+
+
 bool contains_dotdot( const char * const filename )
   {
   for( int i = 0; filename[i]; ++i )
@@ -298,19 +461,19 @@ bool contains_dotdot( const char * const filename )
 int extract_member( const int infd, const Extended & extended,
                     const Tar_header header, const bool keep_damaged )
   {
-  const char * const filename = extended.path.c_str();
+  const char * const filename = extended.path().c_str();
   if( contains_dotdot( filename ) )
     {
     show_file_error( filename, "Contains a '..' component, skipping." );
-    return list_member( infd, extended, header, true );
+    return skip_member( infd, extended );
     }
-  const mode_t mode = strtoul( header + mode_o, 0, 8 );		// 12 bits
-  const time_t mtime = strtoull( header + mtime_o, 0, 8 );	// 33 bits
+  const mode_t mode = parse_octal( header + mode_o, mode_l );	 // 12 bits
+  const time_t mtime = parse_octal( header + mtime_o, mtime_l ); // 33 bits
   const Typeflag typeflag = (Typeflag)header[typeflag_o];
   const bool islink = ( typeflag == tf_link || typeflag == tf_symlink );
   int outfd = -1;
 
-  show_member_name( extended, header, 1 );
+  show_member_name( extended, header, 1, grbuf );
   std::remove( filename );
   make_path( filename );
   switch( typeflag )
@@ -324,13 +487,7 @@ int extract_member( const int infd, const Extended & extended,
     case tf_link:
     case tf_symlink:
       {
-      const char * const linkname = extended.linkpath.c_str();
-/*      if( contains_dotdot( linkname ) )
-        {
-        show_file_error( filename,
-          "Link destination contains a '..' component, skipping." );
-        return list_member( infd, extended, header, false );
-        }*/
+      const char * const linkname = extended.linkpath().c_str();
       const bool hard = typeflag == tf_link;
       if( ( hard && link( linkname, filename ) != 0 ) ||
           ( !hard && symlink( linkname, filename ) != 0 ) )
@@ -352,8 +509,9 @@ int extract_member( const int infd, const Extended & extended,
     case tf_chardev:
     case tf_blockdev:
       {
-      const unsigned dev = makedev( strtoul( header + devmajor_o, 0, 8 ),
-                                    strtoul( header + devminor_o, 0, 8 ) );
+      const unsigned dev =
+        makedev( parse_octal( header + devmajor_o, devmajor_l ),
+                 parse_octal( header + devminor_o, devminor_l ) );
       const int dmode = ( typeflag == tf_chardev ? S_IFCHR : S_IFBLK ) | mode;
       if( mknod( filename, dmode, dev ) != 0 )
         {
@@ -376,8 +534,8 @@ int extract_member( const int infd, const Extended & extended,
       return 2;
     }
 
-  const uid_t uid = (uid_t)strtoul( header + uid_o, 0, 8 );
-  const gid_t gid = (gid_t)strtoul( header + gid_o, 0, 8 );
+  const uid_t uid = (uid_t)parse_octal( header + uid_o, uid_l );
+  const gid_t gid = (gid_t)parse_octal( header + gid_o, gid_l );
   if( !islink && chown( filename, uid, gid ) != 0 &&
       errno != EPERM && errno != EINVAL )
     {
@@ -387,8 +545,8 @@ int extract_member( const int infd, const Extended & extended,
 
   const unsigned bufsize = 32 * header_size;
   uint8_t buf[bufsize];
-  unsigned long long rest = extended.size;
-  const int rem = extended.size % header_size;
+  unsigned long long rest = extended.file_size();
+  const int rem = rest % header_size;
   const int padding = rem ? header_size - rem : 0;
   while( rest > 0 )
     {
@@ -423,25 +581,7 @@ int extract_member( const int infd, const Extended & extended,
   return 0;
   }
 
-
-// Removes any amount of leading "./" and '/' strings.
-const char * remove_leading_slash( const char * const filename )
-  {
-  static bool first_post = true;
-  const char * p = filename;
-
-  while( *p == '/' || ( *p == '.' && p[1] == '/' ) ) ++p;
-  if( p != filename && first_post )
-    {
-    first_post = false;
-    std::string msg( "Removing leading '" );
-    msg.append( filename, p - filename );
-    msg += "' from member names.";
-    show_error( msg.c_str() );
-    }
-  if( *p == 0 ) p = ".";
-  return p;
-  }
+} // end namespace
 
 
 // return true if dir is a parent directory of name
@@ -464,138 +604,179 @@ bool compare_tslash( const char * const name1, const char * const name2 )
   return ( !*p && !*q );
   }
 
-} // end namespace
+namespace {
 
-
-bool Extended::parse( const int infd, const Tar_header header,
-                      const bool permissive )
+bool parse_records( const int infd, Extended & extended,
+                    const Tar_header header, const bool permissive )
   {
-  const unsigned long long edsize = strtoull( header + size_o, 0, 8 );
+  const unsigned long long edsize = parse_octal( header + size_o, size_l );
   const unsigned long long bufsize = round_up( edsize );
   if( bufsize == 0 || edsize == 0 || edsize >= 1ULL << 33 )
     return false;				// overflow or no extended data
   char * const buf = new char[bufsize];		// extended records buffer
-  if( archive_read( infd, (uint8_t *)buf, bufsize ) != 0 ) goto error;
-  for( unsigned long long pos = 0; pos < edsize; )	// parse records
+  const bool ret = ( archive_read( infd, (uint8_t *)buf, bufsize ) == 0 &&
+                     extended.parse( buf, edsize, permissive ) );
+  delete[] buf;
+  return ret;
+  }
+
+} // end namespace
+
+
+/* Returns the number of bytes really read.
+   If (returned value < size) and (errno == 0), means EOF was reached.
+*/
+int readblock( const int fd, uint8_t * const buf, const int size )
+  {
+  int sz = 0;
+  errno = 0;
+  while( sz < size )
     {
-    char * tail;
-    const unsigned long long rsize = strtoull( buf + pos, &tail, 10 );
-    if( rsize == 0 || rsize > edsize - pos || tail[0] != ' ' ||
-        buf[pos+rsize-1] != '\n' ) goto error;
-    ++tail;	// point to keyword
-    // length of (keyword + '=' + value) without the final newline
-    const unsigned long long rest = ( buf + pos + rsize - 1 ) - tail;
-    if( rest > 5 && std::memcmp( tail, "path=", 5 ) == 0 )
-      { if( path.size() && !permissive ) goto error;
-        path.assign( tail + 5, rest - 5 ); }
-    else if( rest > 9 && std::memcmp( tail, "linkpath=", 9 ) == 0 )
-      { if( linkpath.size() && !permissive ) goto error;
-        linkpath.assign( tail + 9, rest - 9 ); }
-    else if( rest > 5 && std::memcmp( tail, "size=", 5 ) == 0 )
-      {
-      if( size != 0 && !permissive ) goto error;
-      size = 0;
-      for( unsigned long long i = 5; i < rest; ++i )
-        {
-        if( tail[i] < '0' || tail[i] > '9' ) goto error;
-        const unsigned long long prev = size;
-        size = size * 10 + ( tail[i] - '0' );
-        if( size < prev ) goto error;		// overflow
-        }
-      if( size < 1ULL << 33 ) goto error;	// size fits in ustar header
-      }
-    else if( rest > 10 && std::memcmp( tail, "GNU.crc32=", 10 ) == 0 )
-      {
-      if( crc_present && !permissive ) goto error;
-      if( rsize != 22 ) goto error;
-      char * t;
-      const uint32_t stored_crc = strtoul( tail + 10, &t, 16 );
-      if( t - tail - 10 != 8 || t[0] != '\n' ) goto error;
-      const uint32_t computed_crc =
-        crc32c.windowed_crc( (const uint8_t *)buf, pos + rsize - 9, edsize );
-      crc_present = true;
-      if( stored_crc != computed_crc ) goto error;
-      }
-    pos += rsize;
+    const int n = read( fd, buf + sz, size - sz );
+    if( n > 0 ) sz += n;
+    else if( n == 0 ) break;				// EOF
+    else if( errno != EINTR ) break;
+    errno = 0;
     }
-  delete[] buf;
-  return true;
-error:
-  delete[] buf;
-  return false;
+  return sz;
+  }
+
+
+/* Returns the number of bytes really written.
+   If (returned value < size), it is always an error.
+*/
+int writeblock( const int fd, const uint8_t * const buf, const int size )
+  {
+  int sz = 0;
+  errno = 0;
+  while( sz < size )
+    {
+    const int n = write( fd, buf + sz, size - sz );
+    if( n > 0 ) sz += n;
+    else if( n < 0 && errno != EINTR ) break;
+    errno = 0;
+    }
+  return sz;
+  }
+
+
+unsigned long long parse_octal( const uint8_t * const ptr, const int size )
+  {
+  unsigned long long result = 0;
+  int i = 0;
+  while( i < size && std::isspace( ptr[i] ) ) ++i;
+  for( ; i < size && ptr[i] >= '0' && ptr[i] <= '7'; ++i )
+    { result <<= 3; result += ptr[i] - '0'; }
+  return result;
   }
 
 
 int decode( const std::string & archive_name, const Arg_parser & parser,
-            const int filenames, const bool keep_damaged, const bool listing,
-            const bool missing_crc, const bool permissive )
+            const int filenames, const int num_workers, const int debug_level,
+            const Program_mode program_mode, const bool ignore_ids,
+            const bool keep_damaged, const bool missing_crc,
+            const bool permissive )
   {
   const int infd = archive_name.size() ?
                    open_instream( archive_name ) : STDIN_FILENO;
   if( infd < 0 ) return 1;
 
-  // execute -C options and mark filenames to be extracted or listed
-  std::vector< bool > name_pending( parser.arguments(), false );
+  // Execute -C options and mark filenames to be compared, extracted or listed.
+  // name_pending is of type char instead of bool to allow concurrent update.
+  std::vector< char > name_pending( parser.arguments(), false );
   for( int i = 0; i < parser.arguments(); ++i )
     {
     const int code = parser.code( i );
-    if( code == 'C' && !listing )
+    if( code == 'C' && program_mode != m_list )
       {
       const char * const dir = parser.argument( i ).c_str();
       if( chdir( dir ) != 0 )
         { show_file_error( dir, "Error changing working directory", errno );
           return 1; }
       }
-    if( !code ) name_pending[i] = true;
+    if( !code && parser.argument( i ).size() ) name_pending[i] = true;
     }
 
-  has_lz_ext =
+  // multi-threaded --list is faster even with 1 thread and 1 file in archive
+  if( program_mode == m_list && num_workers > 0 )
+    {
+    const Lzip_index lzip_index( infd, true, false );	// only regular files
+    const long members = lzip_index.members();
+    if( lzip_index.retval() == 0 && members >= 2 )	// one file + eof
+      {
+      // show_file_error( archive_name.c_str(), "Is compressed seekable" );
+      return list_lz( parser, name_pending, lzip_index, filenames, debug_level,
+                      infd, std::min( (long)num_workers, members ),
+                      missing_crc, permissive );
+      }
+    if( lseek( infd, 0, SEEK_SET ) == 0 && lzip_index.retval() != 0 &&
+        lzip_index.file_size() > 3 * header_size )
+      archive_is_uncompressed_seekable = true;	// unless compressed corrupt
+    }
+
+  has_lz_ext =				// global var for archive_read
     ( archive_name.size() > 3 &&
       archive_name.compare( archive_name.size() - 3, 3, ".lz" ) == 0 ) ||
     ( archive_name.size() > 4 &&
       archive_name.compare( archive_name.size() - 4, 4, ".tlz" ) == 0 );
-  Extended extended;		// metadata from extended records
+  Extended extended;			// metadata from extended records
   int retval = 0;
-  bool prev_extended = false;	// prev header was extended
-  while( true )			// process one member per iteration
+  bool prev_extended = false;		// prev header was extended
+  while( true )				// process one tar member per iteration
     {
-    uint8_t buf[header_size];
-    const int ret = archive_read( infd, buf, header_size );
+    Tar_header header;
+    const int ret = archive_read( infd, header, header_size );
     if( ret == 2 ) return 2;
-    if( ret != 0 || !verify_ustar_chksum( buf ) )
+    if( ret != 0 || !verify_ustar_chksum( header ) )
       {
-      if( ret == 0 && block_is_zero( buf, header_size ) ) break;	// EOF
-      skip_warn(); gretval = 2; continue;
+      if( ret == 0 && block_is_zero( header, header_size ) ) break;	// EOF
+      if( skip_warn() && verbosity >= 2 )
+        std::fprintf( stderr, "ustar chksum = %07o\n", ustar_chksum( header ) );
+      set_error_status( 2 ); continue;
       }
-    skip_warn( true );		// reset warning
+    skip_warn( true );			// reset warning
 
-    const char * const header = (const char *)buf;
     const Typeflag typeflag = (Typeflag)header[typeflag_o];
+    if( typeflag == tf_global )
+      {
+      if( prev_extended )
+        { show_error( "Format violation: global header after extended header." );
+          return 2; }
+      Extended dummy;		// global headers are parsed and ignored
+      if( !parse_records( infd, dummy, header, true ) )
+        { show_error( "Error in global extended records. Skipping to next header." );
+          set_error_status( 2 ); }
+      continue;
+      }
     if( typeflag == tf_extended )
       {
       if( prev_extended && !permissive )
         { show_error( "Format violation: consecutive extended headers found."
-                      /*" Use --permissive."*/, 0, true ); return 2; }
-      if( !extended.parse( infd, header, permissive ) )
+                      /*" Use --permissive.", 0, true*/ ); return 2; }
+      if( !parse_records( infd, extended, header, permissive ) )
         { show_error( "Error in extended records. Skipping to next header." );
-          extended.reset(); gretval = 2; }
-      else if( !extended.crc_present && missing_crc )
+          extended.reset(); set_error_status( 2 ); }
+      else if( !extended.crc_present() && missing_crc )
         { show_error( "Missing CRC in extended records.", 0, true ); return 2; }
       prev_extended = true;
       continue;
       }
     prev_extended = false;
 
-    if( extended.linkpath.empty() )
+    if( extended.linkpath().empty() )	// copy linkpath from ustar header
       {
-      for( int i = 0; i < linkname_l && header[linkname_o+i]; ++i )
-        extended.linkpath += header[linkname_o+i];
-      while( extended.linkpath.size() > 1 &&		// trailing '/'
-             extended.linkpath[extended.linkpath.size()-1] == '/' )
-        extended.linkpath.resize( extended.linkpath.size() - 1 );
+      int len = 0;
+      while( len < linkname_l && header[linkname_o+len] ) ++len;
+      while( len > 1 && header[linkname_o+len-1] == '/' ) --len; // trailing '/'
+      if( len > 0 )
+        {
+        const uint8_t c = header[linkname_o+len]; header[linkname_o+len] = 0;
+        extended.linkpath( (const char *)header + linkname_o );
+        header[linkname_o+len] = c;
+        }
       }
 
-    if( extended.path.empty() )
+    if( extended.path().empty() )		// copy path from ustar header
       {
       char stored_name[prefix_l+1+name_l+1];
       int len = 0;
@@ -606,28 +787,32 @@ int decode( const std::string & archive_name, const Arg_parser & parser,
         { stored_name[len] = header[name_o+i]; ++len; }
       while( len > 0 && stored_name[len-1] == '/' ) --len;	// trailing '/'
       stored_name[len] = 0;
-      extended.path = remove_leading_slash( stored_name );
+      extended.path( remove_leading_dotslash( stored_name ) );
       }
-    const char * const filename = extended.path.c_str();
+    const char * const filename = extended.path().c_str();
 
     bool skip = filenames > 0;
     if( skip )
       for( int i = 0; i < parser.arguments(); ++i )
-        if( parser.code( i ) == 0 )
+        if( !parser.code( i ) && parser.argument( i ).size() )
           {
           const char * const name =
-            remove_leading_slash( parser.argument( i ).c_str() );
+            remove_leading_dotslash( parser.argument( i ).c_str() );
           if( compare_prefix_dir( name, filename ) ||
               compare_tslash( name, filename ) )
             { skip = false; name_pending[i] = false; break; }
           }
 
-    if( extended.size == 0 &&
+    if( extended.file_size() == 0 &&
         ( typeflag == tf_regular || typeflag == tf_hiperf ) )
-      extended.size = strtoull( header + size_o, 0, 8 );
+      extended.file_size( parse_octal( header + size_o, size_l ) );
 
-    if( listing || skip )
-      retval = list_member( infd, extended, header, skip );
+    if( skip )
+      retval = skip_member( infd, extended );
+    else if( program_mode == m_list )
+      retval = list_member( infd, extended, header );
+    else if( program_mode == m_diff )
+      retval = compare_member( infd, extended, header, ignore_ids );
     else
       retval = extract_member( infd, extended, header, keep_damaged );
     extended.reset();
@@ -637,13 +822,10 @@ int decode( const std::string & archive_name, const Arg_parser & parser,
     }
 
   for( int i = 0; i < parser.arguments(); ++i )
-    if( parser.code( i ) == 0 && name_pending[i] )
+    if( !parser.code( i ) && parser.argument( i ).size() && name_pending[i] )
       {
       show_file_error( parser.argument( i ).c_str(), "Not found in archive." );
-      if( gretval < 1 ) gretval = 1;
+      retval = 1;
       }
-  if( !retval && gretval )
-    { show_error( "Exiting with failure status due to previous errors." );
-      retval = gretval; }
-  return retval;
+  return final_exit_status( retval, program_mode != m_diff );
   }
