@@ -32,7 +32,6 @@
 #include <lzlib.h>
 
 #include "arg_parser.h"
-#include "lzip.h"
 #include "lzip_index.h"
 #include "tarlz.h"
 
@@ -76,22 +75,6 @@ int pwriteblock( const int fd, const uint8_t * const buf, const int size,
   }
 
 
-namespace {
-
-// This can be called from any thread, main thread or sub-threads alike,
-// since they all call common helper functions that call cleanup_and_fail()
-// in case of an error.
-//
-void cleanup_and_fail( const int retval = 2 )
-  {
-  // only one thread can delete and exit
-  static pthread_mutex_t mutex = PTHREAD_MUTEX_INITIALIZER;
-
-  pthread_mutex_lock( &mutex );		// ignore errors to avoid loop
-  std::exit( retval );
-  }
-
-
 void xinit_mutex( pthread_mutex_t * const mutex )
   {
   const int errcode = pthread_mutex_init( mutex, 0 );
@@ -162,6 +145,8 @@ void xbroadcast( pthread_cond_t * const cond )
   }
 
 
+namespace {
+
 struct Packet			// member name and metadata or error message
   {
   enum Status { ok, member_done, error };
@@ -263,8 +248,8 @@ public:
     return true;
     }
 
-  // deliver a packet to muxer
-  // if packet.status == Packet::member_done, move to next queue
+  /* Deliver a packet to muxer.
+     If packet.status == Packet::member_done, move to next queue. */
   Packet * deliver_packet()
     {
     Packet * opacket = 0;
@@ -355,8 +340,8 @@ int list_member_lz( LZ_Decoder * const decoder, const int infd,
                     Resizable_buffer & rbuf, const long member_id,
                     const int worker_id, const char ** msg, const bool skip )
   {
-  unsigned long long rest = extended.size;
-  const int rem = extended.size % header_size;
+  unsigned long long rest = extended.file_size();
+  const int rem = rest % header_size;
   const int padding = rem ? header_size - rem : 0;
   const long long data_rest = mdata_end - ( data_pos + rest + padding );
   bool master = false;
@@ -426,9 +411,9 @@ struct Worker_arg
   };
 
 
-       // read lzip members from archive, list their tar members, and
-       // give the produced packets to courier.
-extern "C" void * dworker_l( void * arg )
+/* Read lzip members from archive, list their tar members, and give the
+   packets produced to courier. */
+extern "C" void * tworker( void * arg )
   {
   const Worker_arg & tmp = *(const Worker_arg *)arg;
   const Lzip_index & lzip_index = *tmp.lzip_index;
@@ -442,12 +427,12 @@ extern "C" void * dworker_l( void * arg )
   const int missing_crc = tmp.missing_crc;
   const bool permissive = tmp.permissive;
 
+  Resizable_buffer rbuf( initial_line_length );
   LZ_Decoder * const decoder = LZ_decompress_open();
-  if( !decoder || LZ_decompress_errno( decoder ) != LZ_ok )
-    { show_error( "Not enough memory." ); cleanup_and_fail(); }
+  if( !rbuf.size() || !decoder || LZ_decompress_errno( decoder ) != LZ_ok )
+    { show_error( mem_msg ); cleanup_and_fail(); }
 
   const long long cdata_size = lzip_index.cdata_size();
-  Resizable_buffer rbuf( initial_line_length );
   bool master = false;
   for( long i = worker_id; !master && i < lzip_index.members(); i += num_workers )
     {
@@ -499,7 +484,7 @@ extern "C" void * dworker_l( void * arg )
         {
         if( prev_extended )
           { show_error( "Format violation: global header after extended header." );
-            cleanup_and_fail(); }
+            cleanup_and_fail( 2 ); }
         Extended dummy;		// global headers are parsed and ignored
         const int ret = parse_records_lz( decoder, infd, file_pos, member_end,
                           cdata_size, data_pos, dummy, header, &msg, true );
@@ -527,7 +512,7 @@ extern "C" void * dworker_l( void * arg )
             ret = 2; }
         else ret = parse_records_lz( decoder, infd, file_pos, member_end,
                           cdata_size, data_pos, extended, header, &msg, permissive );
-        if( ret == 0 && !extended.crc_present && missing_crc )
+        if( ret == 0 && !extended.crc_present() && missing_crc )
           { msg = "Missing CRC in extended records."; ret = 2; }
         if( ret != 0 )
           {
@@ -549,16 +534,20 @@ extern "C" void * dworker_l( void * arg )
         }
       prev_extended = false;
 
-      if( extended.linkpath.empty() )	// copy linkpath from ustar header
+      if( extended.linkpath().empty() )	// copy linkpath from ustar header
         {
-        for( int i = 0; i < linkname_l && header[linkname_o+i]; ++i )
-          extended.linkpath += header[linkname_o+i];
-        while( extended.linkpath.size() > 1 &&		// trailing '/'
-               extended.linkpath[extended.linkpath.size()-1] == '/' )
-          extended.linkpath.resize( extended.linkpath.size() - 1 );
+        int len = 0;
+        while( len < linkname_l && header[linkname_o+len] ) ++len;
+        while( len > 1 && header[linkname_o+len-1] == '/' ) --len; // trailing '/'
+        if( len > 0 )
+          {
+          const uint8_t c = header[linkname_o+len]; header[linkname_o+len] = 0;
+          extended.linkpath( (const char *)header + linkname_o );
+          header[linkname_o+len] = c;
+          }
         }
 
-      if( extended.path.empty() )		// copy path from ustar header
+      if( extended.path().empty() )		// copy path from ustar header
         {
         char stored_name[prefix_l+1+name_l+1];
         int len = 0;
@@ -569,25 +558,25 @@ extern "C" void * dworker_l( void * arg )
           { stored_name[len] = header[name_o+i]; ++len; }
         while( len > 0 && stored_name[len-1] == '/' ) --len;	// trailing '/'
         stored_name[len] = 0;
-        extended.path = remove_leading_slash( stored_name );
+        extended.path( remove_leading_dotslash( stored_name ) );
         }
-      const char * const filename = extended.path.c_str();
+      const char * const filename = extended.path().c_str();
 
       bool skip = filenames > 0;
       if( skip )
         for( int i = 0; i < parser.arguments(); ++i )
-          if( parser.code( i ) == 0 )
+          if( !parser.code( i ) && parser.argument( i ).size() )
             {
             const char * const name =
-              remove_leading_slash( parser.argument( i ).c_str() );
+              remove_leading_dotslash( parser.argument( i ).c_str() );
             if( compare_prefix_dir( name, filename ) ||
                 compare_tslash( name, filename ) )
               { skip = false; name_pending[i] = false; break; }
             }
 
-      if( extended.size == 0 &&
+      if( extended.file_size() == 0 &&
           ( typeflag == tf_regular || typeflag == tf_hiperf ) )
-        extended.size = parse_octal( header + size_o, size_l );
+        extended.file_size( parse_octal( header + size_o, size_l ) );
 
       retval = list_member_lz( decoder, infd, file_pos, member_end,
                                cdata_size, data_pos, mdata_end, courier,
@@ -599,7 +588,7 @@ extern "C" void * dworker_l( void * arg )
       else if( retval > 0 )
         { show_error( msg );
           show_error( "Error is not recoverable: exiting now." );
-          cleanup_and_fail(); }
+          cleanup_and_fail( 2 ); }
       }
     }
   if( LZ_decompress_close( decoder ) < 0 )
@@ -614,8 +603,8 @@ done:
   }
 
 
-     // get from courier the processed and sorted packets, and print
-     // the member lines on stdout or the diagnostics on stderr.
+/* Get from courier the processed and sorted packets, and print
+   the member lines on stdout or the diagnostics on stderr. */
 void muxer( Packet_courier & courier )
   {
   while( true )
@@ -624,14 +613,14 @@ void muxer( Packet_courier & courier )
     if( !opacket ) break;	// queue is empty. all workers exited
 
     if( opacket->status == Packet::error )
-      { show_error( opacket->line.c_str() ); cleanup_and_fail(); }
+      { show_error( opacket->line.c_str() ); cleanup_and_fail( 2 ); }
     if( opacket->line.size() )
       { std::fputs( opacket->line.c_str(), stdout );
         std::fflush( stdout ); }
     delete opacket;
     }
   if( !courier.mastership_granted() )	// no worker found EOF blocks
-    { show_error( "Archive ends unexpectedly." ); cleanup_and_fail(); }
+    { show_error( "Archive ends unexpectedly." ); cleanup_and_fail( 2 ); }
   }
 
 } // end namespace
@@ -643,13 +632,15 @@ int list_lz( const Arg_parser & parser, std::vector< char > & name_pending,
              const int debug_level, const int infd, const int num_workers,
              const bool missing_crc, const bool permissive )
   {
-  const int out_slots = 100;
+  const int out_slots = 65536;		// max small files (<=512B) in 64 MiB
+
+  /* If an error happens after any threads have been started, exit must be
+     called before courier goes out of scope. */
   Packet_courier courier( num_workers, out_slots );
 
   Worker_arg * worker_args = new( std::nothrow ) Worker_arg[num_workers];
   pthread_t * worker_threads = new( std::nothrow ) pthread_t[num_workers];
-  if( !worker_args || !worker_threads )
-    { show_error( "Not enough memory." ); cleanup_and_fail(); }
+  if( !worker_args || !worker_threads ) { show_error( mem_msg ); return 1; }
   for( int i = 0; i < num_workers; ++i )
     {
     worker_args[i].lzip_index = &lzip_index;
@@ -663,7 +654,7 @@ int list_lz( const Arg_parser & parser, std::vector< char > & name_pending,
     worker_args[i].missing_crc = missing_crc;
     worker_args[i].permissive = permissive;
     const int errcode =
-      pthread_create( &worker_threads[i], 0, dworker_l, &worker_args[i] );
+      pthread_create( &worker_threads[i], 0, tworker, &worker_args[i] );
     if( errcode )
       { show_error( "Can't create worker threads", errcode ); cleanup_and_fail(); }
     }
@@ -681,7 +672,7 @@ int list_lz( const Arg_parser & parser, std::vector< char > & name_pending,
 
   int retval = 0;
   for( int i = 0; i < parser.arguments(); ++i )
-    if( parser.code( i ) == 0 && name_pending[i] )
+    if( !parser.code( i ) && parser.argument( i ).size() && name_pending[i] )
       {
       show_file_error( parser.argument( i ).c_str(), "Not found in archive." );
       retval = 1;
